"""Rewrite Existing Functions"""

import os
import re
import pdb
import sys
import random
import struct
import builtins
import itertools
import subprocess
from time import time as _time


__all__ = list(globals())


print = lambda *value, file=True: builtins.print(' '.join(map(str, value)) + '\n', end='', file={True: sys.stdout, False: sys.stderr}.get(file, file))
time = lambda start=0: _time() - start
loop = itertools.count
randbytes = lambda n: builtins.bytes(random.randint(0, 255) for i in range(n))
breakpoint = lambda: pdb.set_trace()
popen = lambda cmd: subprocess.Popen(cmd, -1, None, -1, -1, -1, shell=True).stdout
listdir = lambda *paths: os.listdir(os.path.join(*paths))
findall = lambda pattern, string, flags=0: [(m.start(), m.end(), m.group()) for m in re.finditer(pattern, string, flags)]
split = lambda arr, cols: [arr[i:i+cols] for i in range(0, len(arr), cols)]
pack = lambda fmt, values: struct.pack(fmt, *values)
unpack = lambda fmt, string: struct.unpack_from(fmt, string) + (string[struct.calcsize(fmt):],)


def detect(file):
    import chardet
    with builtins.open(file, 'rb') as f:
        b = f.read()
    return chardet.detect(b)['encoding'] or 'utf-8'


def walk(paths='.', exts=''):
    paths = paths if isinstance(paths, (list, tuple)) else [paths]
    exts = [exts] if isinstance(exts, str) else exts
    exts = [ext.lower() for ext in exts]
    for path in paths:
        if os.path.isfile(path):
            if any(path.lower().endswith(ext) for ext in exts):
                yield path
        for root, folders, files in os.walk(path):
            for file in files:
                if any(file.lower().endswith(ext) for ext in exts):
                    yield os.path.join(root, file)


def bytes(data, width=16):
    s = ''.join('\\x%02x' % c for c in data)
    s = "(b'%s')" % "'\n b'".join(s[i: i + width * 4] for i in range(0, len(s), width * 4))
    return s if len(data) > width else s.strip('()')


def memoryview(data, width=16, offset=0):
    s = ' '.join('%02X' % c for c in data)
    return '\n'.join('%08X ' % (i + offset) + s[i * 3: (i + width) * 3] for i in range(0, len(data), width))


__all__ = [k for k in globals() if k not in __all__]



"""Decorators"""

import sys
import time
import traceback
from functools import wraps
from threading import Thread


__all__ = list(globals())


def main(fn):
    import inspect
    if '__main__' == inspect.currentframe().f_back.f_globals['__name__']:
        try:
            if len(sys.argv) == 1:
                fn()
            for v in sys.argv[1:]:
                fn(v)
        except Exception:
            import traceback
            traceback.print_exc()
        input('Press enter to exit: ')
    return fn


def fn2parser(fn):
    import inspect
    import argparse
    if '__main__' != inspect.currentframe().f_back.f_globals['__name__']:
        return fn
    varnames = fn.__code__.co_varnames
    defaults = fn.__defaults__ or ()
    nargs = len(varnames) - len(defaults)
    parser = argparse.ArgumentParser(description=fn.__doc__)
    for i, varname in enumerate(varnames):
        help = None if i < nargs else 'default: ' + repr(defaults[i-nargs])
        parser.add_argument(nargs='?',      dest='pos_' + varname, help=help, metavar=varname)
        parser.add_argument('--' + varname, dest='key_' + varname, help=help, metavar=varname.upper())
    opt = parser.parse_args()
    args = [eval(v) for k, v in vars(opt).items() if k.startswith('p') and v is not None]
    kwargs = {k[4:]: eval(v) for k, v in vars(opt).items() if k.startswith('k') and v is not None}
    fn(*args, **kwargs)
    return fn


def fn2input(fn):
    import inspect
    if '__main__' != inspect.currentframe().f_back.f_globals['__name__']:
        return fn
    varnames = fn.__code__.co_varnames
    defaults = fn.__defaults__ or ()
    nargs = len(varnames) - len(defaults)
    print(fn.__code__.co_name + '(', end='\n' if varnames else '')
    args = [eval(input('  %s = ' % k)) for k in varnames[:nargs]]
    kwargs = {k: eval(input('  %s = %s or ' % (k, repr(v))) or 'v') for k, v in zip(varnames[nargs:], defaults)}
    print(')\n')
    fn(*args, **kwargs)
    return fn


def timeit(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        s = time.time()
        ret = fn(*args, **kwargs)
        e = time.time()
        print('Running <%s>: %.4f' % (fn.__name__, e - s))
        return ret
    return wrapper


def tracer(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        print('Enter function:', fn.__name__)
        if args:
            print('args:', args)
        if kwargs:
            print('kwargs:', kwargs)
        s = time.time()
        ret = fn(*args, **kwargs)
        e = time.time()
        print('Leave function:', e - s)
        return ret
    return wrapper


def protect(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            msg = traceback.format_exc()
            print(msg, file=sys.stderr)
            return msg
    return wrapper


def surround(before=(), after=()):
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            # print(func.__name__) # for test
            [f() for f in before]
            ret = fn(*args, **kwargs)
            [f() for f in after]
            return ret
        return wrapper
    return decorator


def hotkey(key='F12'):
    key = key.lower()
    def decorator(fn):
        def press(key2):
            if key == str(key2).split('.')[-1].lower():
                fn()
        def th():
            import pynput
            with pynput.keyboard.Listener(press) as kl:
                kl.join()
        Thread(target=th).start()
        return fn
    return decorator


def threads(cnt):
    import threading
    counter = threading.BoundedSemaphore(cnt)
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            def th():
                try:
                    fn(*args, **kwargs)
                except:
                    raise
                finally:
                    counter.release()
            counter.acquire()
            threading.Thread(target=th).start()
        return wrapper
    return decorator


__all__ = [k for k in globals() if k not in __all__]


if __name__ == '__main__':
    import random

    @threads(10)
    def delay():
        t = 1 + random.random() / 5
        assert t < 1.19
        time.sleep(t)
        print(t)

    while 1:
        delay()



"""Useful Functions with Built-in Library"""

import os
import html
import time
import traceback
import urllib.parse
import urllib.request


__all__ = list(globals())


# ---------------------------------------------------------------------------
# Debug
# ---------------------------------------------------------------------------


def log(*value, file='log.txt'):
    string = ' '.join(map(str, value)) + '\n'
    header = time.strftime('[%Y-%m-%d %H:%M:%S] ')
    print(string, end='')
    with open(file, 'a', encoding='u8') as f:
        f.write(header + string)


def sudo():
    import sys, ctypes
    if not ctypes.windll.shell32.IsUserAnAdmin():
        args = ' '.join('"' + arg + '"' for arg in sys.argv)
        ok = ctypes.windll.shell32.ShellExecuteW(None, 'runas', sys.executable, args, None, 1) > 32
        exit()


def check(obj, patt='.*', stdout=True):
    import re
    import sys
    import inspect
    patt = re.compile(patt)
    stdout = {True: sys.stdout, False: sys.stderr}.get(stdout, stdout)
    print('obj: %r' % obj, end='')
    for key in sorted(dir(obj)):
        attr = getattr(obj, key)
        try:
            key += str(inspect.signature(attr))
        except (TypeError, ValueError):
            pass
        if patt.fullmatch(key):
            key = '.' + key
            if not callable(attr):
                result = repr(attr)
                key += ' = '
            else:
                try:
                    result = repr(attr())
                    key += ' = '
                except Exception:
                    result = ''
            print('\n' + key, end='')
            print(result, file=stdout, end='')
    print()


def print_paths():
    import os, sys, inspect
    print('\nWORK:\n' + os.getcwd())
    print('\nPATH:\n' + '\n'.join(sys.path))
    print('\nFILE:\n' + inspect.currentframe().f_back.f_globals['__file__'])
    print('\nARGS:\n' + '\n'.join(sys.argv))


def print_lines(lines):
    print('\n'.join(map(str, lines)))


def print_table(table):
    cols = max(len(row) for row in table)
    table = [[str(cell) for cell in row] + [''] * (cols - len(row)) for row in table]
    widths = [max(len(item) for item in column) for column in zip(*table)]
    for row in table:
        line = [item.ljust(width) for item, width in zip(row, widths)]
        print(' | '.join(line))


_last = 0
def progress(*value, interval=1):
    global _last
    now = time.time()
    if interval == 0 or now - _last > interval:
        _last = now
        print(*value)


_fps_n = -1
_fps_t1 = 0
def fps():
    import time
    global _fps_n, _fps_t1
    _fps_n += 1
    fps_t2 = time.time()
    if _fps_t1 == 0:
        _fps_t1 = fps_t2
    if fps_t2 - _fps_t1 > 1:
        print('fps: %.1f' % (_fps_n / (fps_t2 - _fps_t1)))
        _fps_t1 = fps_t2
        _fps_n = 0
    return True


_count_dict = {}
def count(value=None, target=None):
    if value not in _count_dict:
        _count_dict[value] = 0
    if target is None or value == target:
        _count_dict[value] += 1
    return _count_dict[value]


def recent(iterable, max=0):
    history = []
    for item in iterable:
        history.append(item)
        yield history
        if len(history) == max:
            history.pop(0)


def parser2opt(parser, opt='opt'):
    actions = parser._actions[1:]
    w1 = max(len(ac.dest) for ac in actions)
    w2 = max(len(repr(ac.default)) for ac in actions)
    for ac in actions:
        print('%s.%s = %s  # %s' % (opt, ac.dest.ljust(w1), repr(ac.default).ljust(w2), ac.help))


# ---------------------------------------------------------------------------
# File system
# ---------------------------------------------------------------------------


ext = lambda p: os.path.splitext(p)[1].lower()
stem = lambda p: os.path.splitext(os.path.basename(p))[0]
select = lambda p: os.popen('explorer /select, "%s"' % os.path.abspath(p))

file_mtime = lambda p: time.localtime(os.stat(p).st_mtime)[:6]
file_ctime = lambda p: time.localtime(os.stat(p).st_ctime)[:6]
file_utime = lambda p, date: os.utime(p, (time.mktime((tuple(date) + (0,) * 6)[:9]),) * 2)

path_mark = lambda p, mark='.bak': '{0}{2}{1}'.format(*(os.path.splitext(p) + (mark,)))
path_safe = lambda p, repl=None: p.translate({ord(c): urllib.parse.quote_plus(c) if repl is None else repl for c in '\r\n\t\\/:*?"<>|'})
path_split = lambda p: (os.path.dirname(p),) + os.path.splitext(os.path.basename(p))


def path_unique(p, dash='-', start=2):
    root, ext = os.path.splitext(p)
    while os.path.exists(p):
        p = '%s%s%d%s' % (root, dash, start, ext)
        start += 1
    return p


# ---------------------------------------------------------------------------
# Math
# ---------------------------------------------------------------------------


def factorize(num):
    import math
    divisor = 2
    while divisor <= math.sqrt(num):
        if num % divisor == 0:
            print(divisor, '*', end=' ')
            num //= divisor
        else:
            divisor += 1
    print(num)


def primes(max):
    ps = []
    for n in range(2, max):
        if all(n % p for p in ps):
            ps.append(n)
    return ps


# ---------------------------------------------------------------------------
# Web
# ---------------------------------------------------------------------------


escape = html.escape
unescape = html.unescape
quote = urllib.parse.quote_plus  # quote every word include '/'
unquote = urllib.parse.unquote


def urlopen(url, base='', data=None, headers={}, method=None, timeout=10):
    headers = headers or {'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.80 Safari/537.36'}
    url = urllib.parse.urljoin(base, url)
    request = urllib.request.Request(url, data, headers, method=method)
    response = urllib.request.urlopen(request, timeout=timeout)
    return response.read()


# ---------------------------------------------------------------------------
# ...
# ---------------------------------------------------------------------------


def scan(format, string):
    import re
    sp = re.split('(%d|%f|%s)', format, flags=re.I)
    patt = ''.join('(.*)' if i % 2 else re.escape(s) for i, s in enumerate(sp))
    match = re.fullmatch(patt, string)
    fun_map = {'%d': int, '%f': float, '%s': str}
    formats = re.findall('%d|%f|%s', format, flags=re.I)
    return [fun_map[fmt.lower()](s) for fmt, s in zip(formats, match.groups())]


def findpair(s, p1='(', p2=')', st=0):
    n1 = n2 = 0
    for n, c in enumerate(s[st:]):
        n1 += c in p1
        n2 += c in p2
        if n1 and n1 == n2:
            return st + n


def install(path):
    import pip
    pip.main(['install', path])


def input_wait(msg):
    while input(msg + '[y/n]: ').lower() != 'y':
        pass


input_default = lambda msg, default: input('input <%s>, keep <%s> press enter: ' % (msg, default)) or default


class Catch:
    def __init__(self, log='log.txt'):
        self.log = log

    def __enter__(self):
        pass

    def __exit__(self, *args):
        if any(args):
            traceback.print_exc()
            error = traceback.format_exc()
            tt = time.strftime('[%Y-%m-%d %H:%M:%S] ')
            with open(self.log, 'a', encoding='u8') as f:
                f.write(tt + error + '\n')
        return True


catch = Catch()


__all__ = [k for k in globals() if k not in __all__]



"""Console Scripts"""


def run():
    import os
    import sys
    import inspect

    import html
    import urllib.parse
    from . import __version__

    pkg = __import__

    help = lambda: 'Commands: ' + ' '.join(funcs)
    version = lambda: '%s %s from "%s" (python %d.%d.%d)' % ((__package__, __version__, os.path.dirname(__file__)) + sys.version_info[:3])

    ip = lambda: pkg('socket').gethostbyname(pkg('socket').gethostname())
    mac = lambda: '-'.join(pkg('re').findall('..', pkg('uuid').uuid1().hex[-12:].upper()))
    user = lambda: pkg('getpass').getuser()

    crc = lambda p: '0x%04X' % pkg('binascii').crc_hqx(open(p, 'rb').read(), 0)
    md5 = lambda p: pkg('hashlib').md5(open(p, 'rb').read()).hexdigest()
    inv = lambda p: open(p + '.inv', 'wb').write(bytes(255 - b for b in open(p, 'rb').read()))
    create = lambda p: os.path.dirname(p) and os.makedirs(os.path.dirname(p), exist_ok=True) or os.path.basename(p) and open(p, 'w').close()
    delete = lambda p: os.remove(p) if os.path.isfile(p) else pkg('shutil').rmtree(p) if os.path.isdir(p) else None
    detect = lambda p: pkg('chardet').detect(open(p, 'rb').read())['encoding'] or 'utf-8'

    half2hex = lambda num: pkg('struct').pack('>e', float(num)).hex()
    hex2half = lambda hex: pkg('struct').unpack('>e', bytes.fromhex(hex))[0]
    float2hex = lambda num: pkg('struct').pack('>f', float(num)).hex()
    hex2float = lambda hex: pkg('struct').unpack('>f', bytes.fromhex(hex))[0]
    double2hex = lambda num: pkg('struct').pack('>d', float(num)).hex()
    hex2double = lambda hex: pkg('struct').unpack('>d', bytes.fromhex(hex))[0]

    escape = html.escape
    unescape = html.unescape
    quote = urllib.parse.quote_plus
    unquote = urllib.parse.unquote

    funcs = [k for k, v in vars().copy().items() if inspect.isfunction(v)]

    if len(sys.argv) == 1:
        sys.argv.append('help')
    name, *args = sys.argv[1:]
    print(vars()[name](*args))


if __name__ == '__main__':
    run()



"""Lambda Functions"""

import os
import re
import sys
import json
import uuid
import shutil
import socket
import struct
import getpass
import hashlib
import binascii
from threading import Thread

__all__ = list(globals())

# quick function
fake = lambda *v, **kv: None
pause = lambda: input('Press enter to continue: ')
start = lambda fn, *v, **kv: Thread(target=fn, args=v, kwargs=kv).start()
freeze = lambda fn, *v, **kv: (lambda: fn(*v, **kv))

# file function
crc = lambda b: '0x%04X' % binascii.crc_hqx(b if isinstance(b, bytes) else open(b, 'rb').read(), 0)
md5 = lambda b: hashlib.md5(b if isinstance(b, bytes) else open(b, 'rb').read()).hexdigest()
inv = lambda p: open(p + '.inv', 'wb').write(bytes(255 - b for b in open(p, 'rb').read()))
create = lambda p: os.path.dirname(p) and os.makedirs(os.path.dirname(p), exist_ok=True) or os.path.basename(p) and open(p, 'w').close()
delete = lambda p: os.remove(p) if os.path.isfile(p) else shutil.rmtree(p) if os.path.isdir(p) else None

# data resort
unique = lambda arr: sorted(set(arr), key=arr.index)
flatten = lambda arr: sum(arr[1:], arr[0])
reshape = lambda arr, width: list(zip(*[iter(arr)]*width))
transpose = lambda arr: list(zip(*arr))

# string function
join = lambda *s, sp='': sp.join(s)
wash = lambda func, arr: [wash(func, v) for v in arr] if isinstance(arr, (list, tuple)) else func(arr)
dumps = lambda data: json.dumps(data, ensure_ascii=False, indent=2)
indent = lambda width, text: '\n'.join((' ' * width + line)[min(max(0, -width), len(line) - len(line.lstrip())):].rstrip() for line in text.split('\n')[1:-1])
str2dict = lambda s: dict(re.findall(r'(.*?): *(.*)', s))

# sorting
sort_kv  = lambda d, reverse=False: sorted(d.items(), key=lambda item: item[1], reverse=reverse)
sort_key = lambda d, reverse=False: sorted(d, key=d.__getitem__, reverse=reverse)
sort_num = lambda s: [(s, int(n)) for s, n in re.findall(r'(\D+)(\d+)', 'a%s0' % s)]

# number convert
hex2bin = lambda s: bin(int(s, 16))[2:].zfill(len(s) * 4)
bin2hex = lambda s: hex(int(s, 2))[2:].zfill(len(s) // 4)
half2hex = lambda n: struct.pack('>e', n).hex()
hex2half = lambda s: struct.unpack('>e', bytes.fromhex(s))[0]
float2hex = lambda n: struct.pack('>f', n).hex()
hex2float = lambda s: struct.unpack('>f', bytes.fromhex(s))[0]
double2hex = lambda n: struct.pack('>d', n).hex()
hex2double = lambda s: struct.unpack('>d', bytes.fromhex(s))[0]

# pc tools
ip = lambda: socket.gethostbyname(socket.gethostname())
mac = lambda: '-'.join(re.findall('..', uuid.uuid1().hex[-12:].upper()))
user = lambda: getpass.getuser()

# others
tuple2item = lambda item: item if len(item) > 1 else item[0]


__all__ = [k for k in globals() if k not in __all__]



import shutil
import setuptools

pkg = setuptools.find_packages()
pkg_name = pkg[0]

shutil.copy('README.md', pkg_name + '/docs/README.md')
shutil.copy('LICENSE.txt', pkg_name + '/docs/LICENSE.txt')

setuptools.setup(
    name=pkg_name,
    version=__import__(pkg_name).__version__,
    author='Shixian Li (znsoooo)',
    author_email='lsx7@sina.com',
    description='Lite Software eXtension',
    long_description=open('README.md', encoding='u8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/znsoooo/lishixian',
    project_urls={
        'Bug Tracker': 'https://github.com/znsoooo/lishixian/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
    ],
    packages=pkg,

    # install_requires=['pkg_name'],
    # extras_requires={'pkg_name': ['pkg_name']},
    license='MIT License',
    entry_points={'console_scripts': ['{0}={0}.__main__:run'.format(pkg_name)]},
    package_data={'': ['*.*']},
    keywords='lishixian lsx',
)
